/**
IBM SOFTWARE DISCLAIMER 

Java array package (draft 0.2). Copyright (1998), International Business 
Machines Corporation.

Permission to use, copy, modify and distribute this software for any
noncommercial purpose and without fee is hereby granted, provided that
this copyright and permission notice appear on all copies of the
software. The name of the IBM Corporation may not be used in any
advertising or publicity pertaining to the use of the software. IBM
makes no warranty or representations about the suitability of the
software for any purpose.  It is provided "AS IS" without any express
or implied warranty, including the implied warranties of
merchantability, fitness for a particular purpose and non-infringement.
IBM shall not be liable for any direct, indirect, special or
consequential damages resulting from the loss of use, data or projects,
whether in an action of contract or tort, arising out of or in
connection with the use or performance of this software.
*/

package array;

/**
   doubleArray3D = 3-dimensional array of doubles
   <P>
   The doubleArray3D concrete class implements 3-dimensional arrays
   of doubles. 3-dimensional arrays can either be instantiated directly 
   through constructors, or they can be
   generated by extracting three-dimensional sections from 
   other 3- or higher-dimensional arrays.
*/

public final class doubleArray3D extends doubleArray {

    protected double[] data;
    protected int n0, n1, n2;
    protected int w0, w1, w2, w3;

    /**
       Create a 3-dimensional array of a certain shape.
       Elements are 0-valued
       @param n		extent of the array along 0th axis
       @param m		extent of the array along 1st axis
       @param l		extent of the array along 2nd axis
       @exception InvalidArrayShapeException all extents must be nonnegative
    */
    public doubleArray3D(int n, int m, int l) throws InvalidArrayShapeException {

	if (n < 0) throw new InvalidArrayShapeException();
	if (m < 0) throw new InvalidArrayShapeException();
	if (l < 0) throw new InvalidArrayShapeException();
	long size = n*m;
	if (size > Integer.MAX_VALUE) throw new OutOfMemoryError();
	size = size*l;
	if (size > Integer.MAX_VALUE) throw new OutOfMemoryError();
	data = new double[n*m*l];
	if (data == null) throw new InvalidArrayShapeException();
	n0 = n;   n1 = m; n2 = l;
	w0 = m*l; w1 = l; w2 = 1; w3 = 0;
    }

    /**
       Create a 3-dimensional array, with shape and values defined by
       a Java double[][][] array.
       @param a		Java double[][][] array defining shape and values of the array
       @exception InvalidArrayShapeException The Java array must be rectangular
    */
    public doubleArray3D(double[][][] a) throws InvalidArrayShapeException {

	if (a == null) throw new InvalidArrayShapeException();
	int n = a.length;
	int m = a[0].length;
	int l = a[0][0].length;
	long size = n*m;
	if (size > Integer.MAX_VALUE) throw new OutOfMemoryError();
	size = size*l;
	if (size > Integer.MAX_VALUE) throw new OutOfMemoryError();
	data = new double[n*m*l];
	if (data == null) throw new InvalidArrayShapeException();
	for (int i=0; i<n; i++) {
	    if (a[i].length != m) throw new InvalidArrayShapeException();
	    for (int j=0; j<m; j++) {
		if (a[i][j].length != l) throw new InvalidArrayShapeException();
		for (int k=0; k<l; k++) {
		    data[i*m*l+j*l+k] = a[i][j][k];
		}
	    }
	}
	n0 = n;   n1 = m; n2 = l;
	w0 = m*l; w1 = l; w2 = 1; w3 = 0;
    }

    /**
       Create a 3-dimensional array as a section of another array.
       Internal use within the package only.
       A 3-dimensional array can be empty (have 0 elements).
    */
    protected doubleArray3D(double[] a, Range I, Range J, Range K) 
	throws ArrayIndexOutOfBoundsException {

	if (I.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (J.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (K.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (I.max()+J.max()+K.max() >= a.length) throw new ArrayIndexOutOfBoundsException();

	data = a;
	n0 = I.length(); n1 = J.length(); n2 = K.length();
	w0 = I.step();   w1 = J.step();   w2 = K.step();
	w3 = I.first()+J.first()+K.first();
    }

    /**
       Return the rank of the 3-dimensional array, i.e. 3.
    */
    public int rank() {

	return 3;
    }

    /**
       Return the number of elements in the array.
    */
    public int size() {

	return n0*n1*n2;
    }

    /**
       Return the extent of the array along its i-th axis.
       @param i		array axis (0 <= i < rank())
       @exception InvalidArrayAxisException i must be between 0 and 2
    */
    public int size(int i) throws InvalidArrayAxisException {

	if (i < 0) throw new InvalidArrayAxisException();
	if (i > 2) throw new InvalidArrayAxisException();

	int shape = 0;
	if (i == 0) shape = n0;
	if (i == 1) shape = n1;
	if (i == 2) shape = n2;
	return shape;
    }

    /**
       Return the shape of the array as an int[] I of length 3.
       I[0] = size(0)
       I[1] = size(1)
       I[2] = size(2)
    */
    public int[] shape() {

	int[] I = new int[3];
	I[0] = n0;
	I[1] = n1;
	I[2] = n2;

	return I;
    }

    /**
       Return the value of an element of the array.
       @param index	index[i] = index along i-th axis, 0 <= index[i] < size(i)
       @exception InvalidArrayIndexException index[] must be of length 3
    */
    public double get(int[] index)
	throws InvalidArrayIndexException, ArrayIndexOutOfBoundsException {

	if (index.length != 3) throw new InvalidArrayIndexException();
	int i = index[0];
	int j = index[1];
	int k = index[2];
	return get(i,j,k);
    }

    /**
       Set the value of an element of the array.
       @param index	index[i] = index along i-th axis, 0 <= index[i] < size(i)
       @param d		value to which element is set
       @exception InvalidArrayIndexException index[] must be of length 3
    */
    public void set(int [] index, double d) 
	throws InvalidArrayIndexException, ArrayIndexOutOfBoundsException {

	if (index.length != 3) throw new InvalidArrayIndexException();
	int i = index[0];
	int j = index[1];
	int k = index[2];
	set(i,j,k,d);
    }

    /**
       Return the value of an element of the array.
       @param i		index along the 0th axis of the array.
       @param j		index along the 1st axis of the array.
       @param k		index along the 2nd axis of the array.
    */
    public double get(int i, int j, int k) throws ArrayIndexOutOfBoundsException {
	
	if (i < 0) throw new ArrayIndexOutOfBoundsException();
	if (i >= n0) throw new ArrayIndexOutOfBoundsException();
	if (j < 0) throw new ArrayIndexOutOfBoundsException();
	if (j >= n1) throw new ArrayIndexOutOfBoundsException();
	if (k < 0) throw new ArrayIndexOutOfBoundsException();
	if (k >= n2) throw new ArrayIndexOutOfBoundsException();

	return data[i*w0+j*w1+k*w2+w3];
    }

    /**
       Set the value of an element of the array.
       @param i		index along the 0th axis of the array
       @param j		index along the 1st axis of the array
       @param k		index along the 2nd axis of the array
       @param d		value to which element is set
    */
    public void set(int i, int j, int k, double val) throws ArrayIndexOutOfBoundsException {
	
	if (i < 0) throw new ArrayIndexOutOfBoundsException();
	if (i >= n0) throw new ArrayIndexOutOfBoundsException();
	if (j < 0) throw new ArrayIndexOutOfBoundsException();
	if (j >= n1) throw new ArrayIndexOutOfBoundsException();
	if (k < 0) throw new ArrayIndexOutOfBoundsException();
	if (k >= n2) throw new ArrayIndexOutOfBoundsException();

	data[i*w0+j*w1+k*w2+w3] = val;
    }

    /**
       Set the value of an element of the array, no check.
       @param i		index along the 0th axis of the array
       @param j		index along the 1st axis of the array
       @param k		index along the 2nd axis of the array
       @param d		value to which element is set
    */
    public void setf(int i, int j, int k, double val) 
	throws ArrayIndexOutOfBoundsException {
	
	data[i*w0+j*w1+k*w2+w3] = val;
    }

    /**
       Extract a 3-dimensional section from the array.
       @param I		range of indices along 0th axis defining the section
       @param J		range of indices along 1st axis defining the section
       @param K		range of indices along 2nd axis defining the section
       @exception InvalidRangeException	ranges of indices must be within bounds of axes
    */
    public doubleArray3D section(Range I, Range J, Range K) throws InvalidRangeException {

	if (I.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (I.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (J.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (J.max() >= n1) throw new ArrayIndexOutOfBoundsException();
	if (K.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (K.max() >= n2) throw new ArrayIndexOutOfBoundsException();

	Range X = new Range(I.first()*w0+w3,I.last()*w0+w3,I.step()*w0);
	Range Y = new Range(J.first()*w1,   J.last()*w1,   J.step()*w1);
	Range Z = new Range(K.first()*w2,   K.last()*w2,   K.step()*w2);
	return new doubleArray3D(data,X,Y,Z);
    }

    /**
       Extract a 2-dimensional section from the array.
       @param I		range of indices along 0th axis defining the section
       @param J		range of indices along 1st axis defining the section
       @param k		index along 2nd axis defining the section
       @exception InvalidRangeException	ranges of indices must be within bounds of axes
    */
    public doubleArray2D section(Range I, Range J, int k) throws InvalidRangeException {

	if (I.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (I.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (J.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (J.max() >= n1) throw new ArrayIndexOutOfBoundsException();
	if (k < 0) throw new ArrayIndexOutOfBoundsException();
	if (k >= n2) throw new ArrayIndexOutOfBoundsException();

	Range X = new Range(I.first()*w0+w3,I.last()*w0+w3,I.step()*w0);
	Range Y = new Range(J.first()*w1+k*w2,J.last()*w1+k*w2,J.step()*w1);
	return new doubleArray2D(data,X,Y);
    }

    /**
       Extract a 2-dimensional section from the array.
       @param I		range of indices along 0th axis defining the section
       @param j		index along 1st axis defining the section
       @param K		range of indices along 2nd axis defining the section
       @exception InvalidRangeException	ranges of indices must be within bounds of axes
    */
    public doubleArray2D section(Range I, int j, Range K) throws InvalidRangeException {

	if (I.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (I.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (j < 0) throw new ArrayIndexOutOfBoundsException();
	if (j >= n1) throw new ArrayIndexOutOfBoundsException();
	if (K.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (K.max() >= n2) throw new ArrayIndexOutOfBoundsException();

	Range X = new Range(I.first()*w0+w3,I.last()*w0+w3,I.step()*w0);
	Range Z = new Range(j*w1+K.first()*w2,j*w1+K.last()*w2,K.step()*w2);
	return new doubleArray2D(data,X,Z);
    }

    /**
       Extract a 2-dimensional section from the array.
       @param i		index along 0th axis defining the section
       @param J		range of indices along 1st axis defining the section
       @param K		range of indices along 2nd axis defining the section
       @exception InvalidRangeException	ranges of indices must be within bounds of axes
    */
    public doubleArray2D section(int i, Range J, Range K) throws InvalidRangeException {

	if (i < 0) throw new ArrayIndexOutOfBoundsException();
	if (i >= n0) throw new ArrayIndexOutOfBoundsException();
	if (J.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (J.max() >= n1) throw new ArrayIndexOutOfBoundsException();
	if (K.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (K.max() >= n2) throw new ArrayIndexOutOfBoundsException();

	Range Y = new Range(J.first()*w1+w3,J.last()*w1+w3,J.step()*w1);
	Range Z = new Range(i*w0+K.first()*w2,i*w0+K.last()*w2,K.step()*w2);
	return new doubleArray2D(data,Y,Z);
    }

    /**
       Extract a 1-dimensional section from the array.
       @param I		range of indices along 0th axis defining the section
       @param j		index along 1st axis defining the section
       @param k		index along 2nd axis defining the section
       @exception InvalidRangeException	range of indices must be within bounds of 0th axis
    */
    public doubleArray1D section(Range I, int j, int k) throws InvalidRangeException {

	if (I.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (I.max() >= n0) throw new ArrayIndexOutOfBoundsException();
	if (j < 0) throw new ArrayIndexOutOfBoundsException();
	if (j >= n1) throw new ArrayIndexOutOfBoundsException();
	if (k < 0) throw new ArrayIndexOutOfBoundsException();
	if (k >= n2) throw new ArrayIndexOutOfBoundsException();

	Range X = new Range(I.first()*w0+j*w1+k*w2+w3,
			    I.last()*w0+j*w1+k*w2+w3,
			    I.step()*w0);
 	return new doubleArray1D(data,X);
    }

    /**
       Extract a 1-dimensional section from the array.
       @param i		index along 0th axis defining the section
       @param J		range of indices along 1st axis defining the section
       @param k		index along 2nd axis defining the section
       @exception InvalidRangeException	range of indices must be within bounds of 1st axis
    */
    public doubleArray1D section(int i, Range J, int k) throws InvalidRangeException {

	if (i < 0) throw new ArrayIndexOutOfBoundsException();
	if (i >= n0) throw new ArrayIndexOutOfBoundsException();
	if (J.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (J.max() >= n1) throw new ArrayIndexOutOfBoundsException();
	if (k < 0) throw new ArrayIndexOutOfBoundsException();
	if (k >= n2) throw new ArrayIndexOutOfBoundsException();

	Range Y = new Range(i*w0+J.first()*w1+k*w2+w3,
			    i*w0+J.last()*w1+k*w2+w3,
			    J.step()*w1);
 	return new doubleArray1D(data,Y);
    }

    /**
       Extract a 1-dimensional section from the array.
       @param i		index along 0th axis defining the section
       @param j		index along 1st axis defining the section
       @param K		range of indices along 2nd axis defining the section
       @exception InvalidRangeException	range of indices must be within bounds of 2nd axis
    */
    public doubleArray1D section(int i,int j, Range K) throws InvalidRangeException {

	if (i < 0) throw new ArrayIndexOutOfBoundsException();
	if (i >= n0) throw new ArrayIndexOutOfBoundsException();
	if (j < 0) throw new ArrayIndexOutOfBoundsException();
	if (j >= n1) throw new ArrayIndexOutOfBoundsException();
	if (K.min() < 0) throw new ArrayIndexOutOfBoundsException();
	if (K.max() >= n2) throw new ArrayIndexOutOfBoundsException();

	Range Z = new Range(i*w0+j*w1+K.first()*w2+w3,
			    i*w0+j*w1+K.last()*w2+w3,
			    K.step()*w2);
 	return new doubleArray1D(data,Z);
    }

    /**
       Extract a 0-dimensional section from the array.
       @param i		index along 0th axis defining the section.
       @param j		index along 1st axis defining the section.
       @param k		index along 2nd axis defining the section.
    */
    public doubleArray0D section(int i, int j, int k) {
	
	if (i < 0) throw new ArrayIndexOutOfBoundsException();
	if (i >= n0) throw new ArrayIndexOutOfBoundsException();
	if (j < 0) throw new ArrayIndexOutOfBoundsException();
	if (j >= n1) throw new ArrayIndexOutOfBoundsException();
	if (k < 0) throw new ArrayIndexOutOfBoundsException();
	if (k >= n2) throw new ArrayIndexOutOfBoundsException();

	return new doubleArray0D(data,i*w0+j*w1+k*w2+w3);
    }

    /**
       Assign a scalar value to all elements of the array.
       @param d		value to be assigned
    */
    public void assign(double d) {

	try {
	    int n = size(0);
	    int m = size(1);
	    int l = size(2);
	    for (int i=0; i<n; i++) {
		for (int j=0; j<m; j++) {
		    for (int k=0; k<l; k++) {
			set(i,j,k,d);
		    }
		}
	    }
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Unexpected internal error");
	    System.exit(-1);
	}
    }

    /**
       Assign the values from another three-dimensional array to this array.
       @param a		array with source values
       @exception NonconformingArrayException arrays must be of same shape
    */
    public void assign(doubleArray3D a) 
	throws NonconformingArrayException {

	try {
	    int n = size(0);
	    int m = size(1);
	    int l = size(2);
	    
	    if (n != a.size(0)) throw new NonconformingArrayException();
	    if (m != a.size(1)) throw new NonconformingArrayException();
	    if (l != a.size(2)) throw new NonconformingArrayException();

	    for (int i=0; i<n; i++) {
		for (int j=0; j<m; j++) {
		    for (int k=0; k<l; k++) {
			set(i,j,k,a.get(i,j,k));
		    }
		}
	    }
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Unexpected internal error");
	    System.exit(-1);
	}
    }
}


    
	
